package com.experimental.irrigator

import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothDevice.DEVICE_TYPE_CLASSIC
import android.bluetooth.BluetoothDevice.DEVICE_TYPE_DUAL
import android.bluetooth.BluetoothSocket
import android.content.Intent
import androidx.appcompat.app.AppCompatActivity
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.os.Message
import android.util.Log
import android.view.View
import android.widget.Switch
import android.widget.TextView
import android.widget.Toast
import android.widget.ToggleButton
import androidx.appcompat.widget.SwitchCompat
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers.Main
import kotlinx.coroutines.Job
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import java.io.IOException
import java.util.*

private val TAG : String = " -- MainActivity: "

private val STATCALL_BYTE : Byte = 0xa1.toByte()
private val FILLREQ_BYTE : Byte = 0xa2.toByte()

class MainActivity : AppCompatActivity() {

    private var mDeviceBonded : Boolean = false
    private var mBlueService : BlueService? = null
    private var mBlueConnected : Boolean = false
    private var mConnectThread : ConnectThread? = null
    private var mBlueCommsManager : BlueCommsManager? = null
    private var moistureDataView : TextView? = null
    private var fillDataView : TextView? = null
    private var tempDataView : TextView? = null
    private var statCallQueued : Boolean = false
    private var fillReqQueued : Boolean = false
    private var blueCallJob : Job? = null
    private var automated : Boolean = false
    private var autoJob : Job? = null
    private var toggleView : Switch? = null

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.activity_main)
        moistureDataView = findViewById<TextView>(R.id.dataView_Moisture)
        fillDataView = findViewById<TextView>(R.id.dataView_Fill)
        tempDataView = findViewById<TextView>(R.id.dataView_Temp)
        toggleView = findViewById(R.id.automate_switch)
        toggleView?.setOnCheckedChangeListener { _, isChecked ->
            if (isChecked) {
                // The toggle is enabled
                automated = true
                setAutomated()
            } else {
                // The toggle is disabled
                automated = false
                setAutomated()
            }
        }

    }

    override fun onStart() {
        super.onStart()

        launchBlueCallJob()
        setAutomated()
    }

    override fun onResume() {
        super.onResume()

        launchBlueCallJob()
        setAutomated()
    }

    override fun onPause() {
        super.onPause()
        autoJob?.cancel()
        blueCallJob?.cancel()
        mConnectThread?.cancel()
    }

    override fun onDestroy() {
        super.onDestroy()
        autoJob?.cancel()
        blueCallJob?.cancel()
        mConnectThread?.cancel()
    }

    fun setAutomated() {
        if (automated) {
            if (autoJob != null) {
                if (!autoJob!!.isActive) launchAutoJob()
            } else launchAutoJob()
        } else {
            autoJob?.cancel()
        }
    }

    fun launchAutoJob() {
        Log.d(TAG, "launching autoJob")
        autoJob = CoroutineScope(Main).launch {
            while(true) {
                delay(1000)
                Log.d( TAG, "autoJob queuing statCall")
                statCallQueued = true
            }
        }
    }

    fun launchBlueCallJob() {
        Log.d(TAG, "launching blueCallJob")
        if (blueCallJob != null) { if (blueCallJob!!.isActive) blueCallJob?.cancel() }
        blueCallJob = CoroutineScope(Main).launch {
            while(true) {
                delay(100)
                if (mConnectThread != null) {
                    if (mConnectThread!!.isAlive) continue
                }
                if (!mBlueConnected) {
                    if (statCallQueued) {
                        Log.d(TAG, "starting queued statcall")
                        initiateCommsThreads(STATCALL_BYTE)
                        statCallQueued = false
                    } else if (fillReqQueued) {
                        Log.d(TAG, "starting queued fillrequest")
                        initiateCommsThreads(FILLREQ_BYTE)
                        fillReqQueued = false
                    }
                }
            }
        }
    }

    fun onButtonClick( view : View ) {
        Log.d( TAG, "get stats button pressed")
        Log.d( TAG, "queuing statCall")
        statCallQueued = true
    }

    fun onButtonClick2( view : View ) {
        Log.d( TAG, "reset fill button pressed")
        Log.d( TAG, "queuing fillRequest")
        fillReqQueued = true
    }

    fun setMoisture( value : UInt ) {
        moistureDataView?.text = value.toString()+" %"
    }
    fun setFill( value : UInt ) {
        fillDataView?.text = value.toString()+" %"
    }
    fun setTemp( value : UInt ) {
        tempDataView?.text = value.toString()+" °C"
    }

    fun initiateCommsThreads( call : Byte ) {
        mBlueCommsManager = BlueCommsManager( call )
        if ( mConnectThread != null )  {
            shortToast( "cancelling Connection .." )
            Log.d( TAG, "cancelling connect Thread" )
            mConnectThread!!.cancel()
            mBlueConnected = false
            return
        }
        val adapter : BluetoothAdapter? = getBlueAdapter()
         if ( adapter != null ) {
             val device : BluetoothDevice? = getBlueDevice( adapter )
             if ( device != null ) {
                 mConnectThread = ConnectThread( device )
                 Log.d( TAG, "starting BT connect Thread")
                 mConnectThread?.start()
             }
         }
    }

    private fun getBlueAdapter() : BluetoothAdapter?  {
        val bluetoothAdapter: BluetoothAdapter? = BluetoothAdapter.getDefaultAdapter()
        if (bluetoothAdapter == null) {
            // Device doesn't support Bluetooth
            shortToast( "no Bluetooth adapter ..")
            return null
        }
        if (bluetoothAdapter.isEnabled == false) {
            //val enableBtIntent = Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE)
            //startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT)
            shortToast( "Bluetooth is disabled .." )
            return null
        }
        return bluetoothAdapter
    }

    private fun getBlueDevice( adapter : BluetoothAdapter ) : BluetoothDevice?  {
        val pairedDevices: Set<BluetoothDevice>? = adapter.bondedDevices
        pairedDevices?.forEach { device ->
            //val deviceName = device.name
            //val deviceHardwareAddress = device.address // MAC address
            if ( device.name == "IPAT0.1") {
                if (device.type == DEVICE_TYPE_DUAL) {
                    Log.d(TAG, "bonded dual BT device " + device.name)
                    mDeviceBonded = true
                    return device
                } else if (device.type == DEVICE_TYPE_CLASSIC) {
                    Log.d(TAG, "bonded classic BT device " + device.name)
                    mDeviceBonded = true
                    return device
                }
            }
        }
        mDeviceBonded = false
        shortToast("Device not bonded ..")
        Log.d( TAG, "no device bonded ")
        return null
    }

    private fun shortToast( msg : String ) {
        Toast.makeText( applicationContext, msg, Toast.LENGTH_SHORT ).show()
    }

    private fun manageMyConnectedSocket( socket : BluetoothSocket ) {
        val blueHandler : Handler = object : Handler(Looper.getMainLooper()) {
            override fun handleMessage(msg : Message) {
                Log.d(TAG,"incoming blueEvent")
                when (msg.what) {
                    MESSAGE_READ -> {
                        //msg.obj.
                        Log.d(TAG, "BT RX message: " + (msg.obj as ByteArray).toString() )
                        mBlueCommsManager?.receivedMsg(msg)
                    }
                    MESSAGE_WRITE   ->  {
                        Log.d(TAG, "BT TX message sent")
                        mBlueCommsManager?.sendConfirmed()
                    }
                    MESSAGE_TOAST   ->  {
                        Log.d(TAG, "BT toast")
                        mBlueCommsManager?.sendConfirmed(failed = true)
                    }
                    MESSAGE_READY   ->  {
                        Log.d(TAG, "BT comms threads ready" )
                        mBlueCommsManager?.sendRequest()
                    }
                    MESSAGE_CLOSED  ->  {
                        if ( mConnectThread == null ) {
                            Log.d( TAG, "received confirmation of BlueConnection end")
                            if ( mConnectThread != null ) {
                                if ( mConnectThread!!.isAlive == false )
                                    Log.d( TAG, "communications threads confirmed dead")
                            } else Log.d( TAG, "thread deaths not confirmed")
                        } else {
                            Log.d(TAG, "received BlueConnection end notice" )
                            mBlueConnected = false
                            if ( mConnectThread != null ) {
                                if ( mConnectThread!!.isAlive == false )
                                    Log.d( TAG, "communications threads confirmed dead")
                            } else Log.d( TAG, "thread deaths not confirmed")
                            mConnectThread = null
                        }
                    }
                }
            }
        }
        //val blueService : mBlueService =
        Log.d(TAG,"starting blueService")
        if (mBlueService != null) mBlueService?.cancel()
        mBlueService = BlueService(blueHandler)
        mBlueService?.communicate(socket)
    }

    private inner class ConnectThread( device: BluetoothDevice ) : Thread() {
        val MY_UUID : UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB")
        private val mBLueAdapter : BluetoothAdapter? = getBlueAdapter()
        private val mmSocket: BluetoothSocket? by lazy(LazyThreadSafetyMode.NONE) {
            device.createRfcommSocketToServiceRecord(MY_UUID)
        }

        public override fun run() {
            // Cancel discovery because it otherwise slows down the connection.
            //mBLueAdapter?.cancelDiscovery()

            mmSocket?.use { socket ->
                // Connect to the remote device through the socket. This call blocks
                // until it succeeds or throws an exception.
                try { socket.connect() } catch (e: IOException) {
                    Log.e(TAG, "Could not connect the client socket", e)
                }

                // The connection attempt succeeded. Perform work associated with
                // the connection in a separate thread.
                Log.d(TAG, "BT connection succeeded")
                mBlueConnected = true
                manageMyConnectedSocket(socket)
            }
        }

        // Closes the client socket and causes the thread to finish.
        fun cancel() {
            try {
                mmSocket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "Could not close the client socket", e)
            }
        }
    }

    private inner class BlueCommsManager( commsTask : Byte ) {
        private val WAITING1 : Int = 1
        private val WAITING2 : Int = 2
        private val WAITING3 : Int = 3
        private val AWAIT_CONFIRM : Int = 4
        private val TIMEOUTLIMIT : Long = 200
        private val mCommsTask : Byte = commsTask
        private var mCommsState : Int = 0;

        public fun sendRequest() {
            mBlueService?.sendByte( mCommsTask )
        }
        public fun sendConfirmed(failed : Boolean = false) {
            if (failed) {
                Log.d(TAG, "BT request failed, aborting ..")
                /*if (mCommsTask == FILLREQ_BYTE)*/ shortToast("Connection failed ..")
                if (automated) toggleView?.toggle()

                commsCancelation()
            } else if ( mCommsState == 0 ) {
                Log.d(TAG, "confirmed sent message, configuring timeout ..")
                if (mCommsTask == STATCALL_BYTE) mCommsState = WAITING1
                else if (mCommsTask == FILLREQ_BYTE) mCommsState = AWAIT_CONFIRM
                setTimeout(TIMEOUTLIMIT)
            } else {
                Log.d( TAG, "Comms Management Error")
                commsCancelation()
            }
        }
        public fun receivedMsg( msg : Message ) {
            when (mCommsState) {
                    WAITING1    ->  {
                    //mCommsState = WAITING2
                    readData(msg)
                }   WAITING2    ->  {
                    //mCommsState = WAITING3
                    readData(msg)
                }   WAITING3    ->  {
                    readData(msg)
                    //commsCancelation()
                }   AWAIT_CONFIRM    ->  {
                    verifyFillReset(msg)
                }   else    ->  {
                    Log.d( TAG, "Comms Management or Data Error")
                    commsCancelation()
                }
            }
        }
        private fun readData(msg : Message) {
            val data : ByteArray = msg.obj as ByteArray
            val remaining : Int = 3-(mCommsState-1)
            val received = msg.arg1
            var current : Int = mCommsState
            if ( (remaining -received) >= 0 ) {
                repeat(received) { i   ->
                    val byte : Byte = data.get(i)
                    if ( current == 1 ) {
                        Log.d(TAG, "Moisture received")
                        setMoisture(byte.toUInt())
                    } else if ( current == 2 ) {
                        Log.d(TAG, "Fill received")
                        setFill(byte.toUInt())
                    } else if ( current == 3 ) {
                        Log.d(TAG, "Temperature received")
                        setTemp(byte.toUInt())
                    }
                    current++
                }
            } else {
                Log.d(TAG, "WARNING: to much data received, aborting!")
                commsCancelation()
                return
            }
            if ( current > 3 ) {
                Log.d(TAG, "data complete, disconnecting ..")
                commsCancelation()
                return
            } else {
                Log.d(TAG, "waiting for remaining " +(4-current).toString() +" bytes")
                mCommsState = current
            }
        }
        private fun verifyFillReset(msg : Message) {
            Log.d(TAG, "Verifying FillRequest confirmation ..")
            val received = msg.arg1
            val data : ByteArray = msg.obj as ByteArray
            val byte : Byte = data.get(0)
            if ((received == 1) and (byte == FILLREQ_BYTE)) {
                Log.d(TAG, "Fill reset was confirmed")
                shortToast("Fill reset successfull ..")
            } else {
                Log.d(TAG, "Fill reset not confirmed")
                shortToast("Fill reset failed ..")
            }
            commsCancelation()
            return

        }
        private fun commsCancelation() {
            //if ( mConnectThread != null ) {
                //shortToast("cancelling Connection ..")
                Log.d(TAG, "cancelling connect Thread")
                mConnectThread?.cancel()
                mBlueConnected = false
            //}
        }
        private fun setTimeout(mils : Long) {
            CoroutineScope(Main).launch {
                delay(mils)
                if (mBlueConnected) {
                    Log.d(TAG, "connection timed out after " +mils.toString() +" ms")
                    commsCancelation()
                }
            }
        }
    }
}